/*
  Telemetriesender für Endstufen und Transvertereinheiten
  Kommunikation über MODBUS/RTU

***************************************************************************************************************
 *
 *  This program is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License
 *
 *                            http://creativecommons.org/licenses/by-nc-sa/3.0/
 *
 *                        http://creativecommons.org/licenses/by-nc-sa/3.0/legalcode
 *
 *
 ***************************************************************************************************************

  V00.20 2025-02-11-DL1SUZ - Uwe
  Email: dl1suz@darc.de

  Hardwarebasis: Arduino NANO V3 (ATMEL 328P)
  PCB-Layout DL1SUZ

  Die Software wird wie sie ist zur Verfügung gestellt.

  Die Modbus-Kommunikation wird komplett über die die Holding-Register abgewickelt
  40001:           DO[0] - DO[15]  - 16 Einzelbits für Kommunikation vom Master zum Slave
                 reseviert
  40002:           DI[0] - DI[15]  - 16 Einzelbits für Kommunikation vom Slave zum Master
                 reseviert
  40003/40004:     float_reverse:   power_fwd
  40005/40006:     float_reverse:   power_ref
  40007/40008:     float_reverse:   power_drv
  40009/40010:     float_reverse:   U1 (28V/48V)
  40011/40012:     float_reverse:   I1 (ACS712)
  40013/40014:     float_reverse:   U2 (12V)
  40015/40016:     float_reverse:   SWR
  40017/40018:     float_reverse:   temperature
  40017/40018:     float_reverse:   res


*/
//Bibliotheken
#include <ModbusRtu.h>             //"Modbus-Master-Slave-for-Arduino-master" von smarmengol 
//https://github.com/smarmengol/Modbus-Master-Slave-for-Arduino
#include <OneWire.h>
#include <DallasTemperature.h>

//Vereinbarungen für den MODBUS/RTU-Slave
#define TXEN 2                      //PIN D2 : tx-enaable für RS485
uint8_t idSlave = 1;                //Slave Nummer eintragen
uint16_t baudrate = 38400;
uint16_t au16data[20];              // data array for modbus network sharing
uint8_t state;                      // machine state
unsigned long u32wait;
// Instanz erzeugen
Modbus slave(idSlave, Serial, TXEN);  // this is slave and RS-485

//Vereinbarungen für die Temperaturmessung
#define ONE_WIRE_BUS 12    //  Data wire is plugged into pin 12 on the Arduino 
#define TEMPERATURE_PRECISION 13
OneWire oneWire(ONE_WIRE_BUS); // Setup a oneWire instance to communicate with any OneWire devices
DallasTemperature sensors(&oneWire); // Pass oneWire reference to Dallas Temperature
float temperature;

//Vereinbarungen für die Strommessung
#define ACS_sensivity 0.066  //ACS712 +-30A 


// Bereich für gemeinsamen Speicherzugriff vereinbaren (für Modbus-Format float_revers")
union memory
{
  float floats[9];
  word words [18];
};
union memory values;  //Instanz für den Speicher anlegen

// ************************************************************** variables to be changed for calibration *********
// use   https://valdivia.staff.jade-hs.de/interpol.html                                                          *
//       for calculation coefficients          formula used:   y = a +b*x + c*x*x                                 *
//                                                                                                                *
float    a_drv = 0 , b_drv = 0.227, c_drv = 2.01;                   // coefficients of polygon approximation DRV  *
float    a_fwd = 0.504, b_fwd = -2.718, c_fwd = 1.948;              // coefficients of polygon approximation FWD  *
float    a_ref = 0.204, b_ref = -2.718, c_ref = 1.948;              // coefficients of polygon approximation REF  *
float    scc_drv = 1.0, scc_fwd = 1.0, scc_ref =  1.0 ;             // Shift_Calibration_Curve 0.93=7%> 1.07=7%<  *
float    power_drv_min = 0.1;                                       // minimal power for calculation and display  *
byte     power_fwd_min = 1, power_ref_min = 1 , power_drv_max = 1;  // min. power for calc. and display           *
float    voltage_fwd,voltage_ref, voltage_drv;                      //voltages from coupler                       *
float    voltage_fwd_avg,voltage_ref_avg,voltage_drv_avg;           //sum_voltages for average                    * 
byte     iii = 0;                                                   //                                            *
// ****************************************************************************************************************

//Vereinbarungen für das E/A-Handling
//Analog pins
#define IN_FWD    A0  // analog input for forward voltage
#define IN_REF    A1  // analog input for reflected voltage
#define IN_50_U   A2  // analog input 50VDC - Spannungsmessung
#define IN_50_I   A3  // analog input 50VDC - Strommessung
#define IN_15_U   A6  // analog input 15VDC - Spannungsmessung
//Digital pins
#define DI0       4  // Digitaleingang 0 - Pin D4
#define DI1       5  // Digitaleingang 1 - Pin D5
#define DI2       6  // Digitaleingang 2 - Pin D6
#define DI3       7  // Digitaleingang 3 - Pin D7
#define DO0       8  // Digitalausgang 0 - Pin D8
#define DO1       9  // Digitalausgang 1 - Pin D9
#define DO2       10  // Digitalausgang 2 - Pin D10
#define DO3       11  // Digitalausgang 3 - Pin D11
#define LED       13  // interne LED      - Pin D13


//Variablenvereinbarung für das E/A-Handling
float  volt_15 = 0.0;                        // 15v volt supply - voltage
float  volt_50 = 0.0;                        // 50v volt supply - voltage
float  amp_50  = 0.0;                        // 50v volt supply - current
float  power_fwd = 0.0;                      // power forward (watts)
float  power_ref = 0.0;                      // power reflected (watts)
float  power_drv = 0.0;                      // power drive (watts)
float  SWR = 0.0;                            // SWR
float  Pratio = 0.0;                         // Power ratio P forward / P refl
int coilValue = 0;
int inputValue = 0;

unsigned long time; // für Temperaturmessung

void setup() {
  //Vorbereiten der I/O-Pins
  pinMode(DI0, INPUT); //RX/TX
  pinMode(DI1, INPUT_PULLUP);
  pinMode(DI2, INPUT_PULLUP);
  pinMode(DI3, INPUT_PULLUP);
  pinMode(DO0, OUTPUT);
  pinMode(DO1, OUTPUT);
  pinMode(DO2, OUTPUT);
  pinMode(DO3, OUTPUT);
  pinMode(LED, OUTPUT);
  //----------------  MODBUS-Slave  --------------------------
  //Initialisierung des Modbus-Slave
  Serial.begin(baudrate);
  slave.start();
  u32wait = millis() + 100;
  digitalWrite(LED, HIGH);
  //-------------  Temperaturmessung  -------------------------
  sensors.begin();   // Start up the library

  //--------------- E/A-Handling -----------------------------
  // Verdrahtung für MKU13G4 DO0 = PIN24, DO1 = PIN 26
  // RX 2300MHz:  PIN24 = DO0: LOW  | PIN26 = DO1: LOW
  // RX 2304MHz:  PIN24 = DO0: HIGH | PIN26 = DO1: LOW
  // RX 2320MHz:  PIN24 = DO0: LOW  | PIN26 = DO1: HIGH
  // RX 2400MHz:  PIN24 = DO0: HIGH | PIN26 = DO1: HIGH
  // RX-Voreinstellung 2320MHz
  digitalWrite (DO0, LOW);
  digitalWrite (DO1, HIGH);
  // Verdrahtung für MKU13G4 DO2 = PIN20, PIN22 = open
  // TX 2320MHz:  PIN20 = DO2: LOW
  // TX 2400MHz:  PIN20 = DO2: HIGH
  digitalWrite (DO2, LOW);
  au16data[0] = 0x0005;          //Vorbelegung der Variablen au16data[0] für RX/TX 2320MHz
}
//---------------- Beginn des Hauptprogrammes ----------------------
void loop() {
  modbus_slave();   //MODBUS aufrufen
  read_temperature();  //Temperatur lesen des DS18B20 lesen
  read_analog_values();
  value_mapping();
  read_digital_inputs();
  write_digital_outputs();
}
//----------------  Ende des Hauptprogrammes ------------------------
//========================================================================================================

//---------------- Analalogwerte lesen ---------------------------------------------------
void read_analog_values()
{
  //  amp_50  = (analogRead(IN_50_I))*0.376/1023;               // ACS715: Messbereich 0-30A, 133mV/A
  amp_50  = (analogRead(IN_50_I) - 512) * ACS_sensivity / 1023; //ACS712 - Empfindlichkeit lt. ACS_sensivity
  //  amp_50  = (analogRead(IN_50_I)) * 27.5 / 1023;               //BTS 50085 mit Spannungsteiler 2k/1k
  volt_50 = (analogRead(IN_50_U)) * 55.0 / 1023;               //Abgleich
  volt_15 = (analogRead(IN_15_U)) * 20.0 / 1023;               //Abgleich

  voltage_fwd_avg = 0;
  voltage_ref_avg = 0;

  byte xx= 20;
  for(iii=0; iii<xx; iii++)                             //Durchschnitt aus x Messungen bilden
  {
  voltage_fwd = analogRead(IN_FWD);       
  voltage_ref = analogRead(IN_REF);   

  voltage_fwd_avg = voltage_fwd_avg + voltage_fwd; 
  voltage_ref_avg = voltage_ref_avg + voltage_ref;
  } 
  voltage_fwd = voltage_fwd_avg/xx ;                                         // use peak voltage for processing
  voltage_ref = voltage_ref_avg/xx ;

  voltage_fwd = voltage_fwd*14.0/818*scc_fwd;                                //Normalisieren des Wertes 
  voltage_ref = voltage_ref*14.0/818*scc_ref;   

  power_fwd = a_fwd + b_fwd*voltage_fwd + c_fwd*voltage_fwd*voltage_fwd;    // polygon approximatiion equation fwd
    if(power_fwd < 0) power_fwd = 0;                                          // avoid negativ values    
  power_ref = a_ref + b_ref*voltage_ref + c_ref*voltage_ref*voltage_ref;    // polygon approximatiion equation ref
    if(power_ref < 0) power_ref = 0; 
  power_drv = 0.0 ;

  
  //Berechnung SWR
  float reflexion_coefficient = sqrt(power_ref / power_fwd);
  if (reflexion_coefficient <= 0.8) {
    SWR = (1 + reflexion_coefficient) / (1 - reflexion_coefficient);
  }
  else SWR = -1;

}
//---------------- Temeperaurmessung mit DS18B20 One-wire -----------------------------
void read_temperature()
{
  if ((millis() - time) > 5000)                                         // The sensor is only read out every 5s
  { sensors.requestTemperatures();                                      // because the process takes 750ms
    temperature = sensors.getTempCByIndex(0);                           // in high resolution mode.
    time = millis();
  }
}

//---------------- Digitale Ausgänge schreiben ----------------------------------------
void write_digital_outputs()
{
  bool bitResult = bitRead(au16data[0], 0) ;
  if (bitResult == 1) digitalWrite(DO0, LOW);
  else digitalWrite(DO0, HIGH);
  bitResult = bitRead(au16data[0], 1) ;
  if (bitResult == 1) digitalWrite(DO1, LOW);
  else digitalWrite(DO1, HIGH);
  bitResult = bitRead(au16data[0], 2) ;
  if (bitResult == 1) digitalWrite(DO2, LOW);
  else digitalWrite(DO2, HIGH);
  bitResult = bitRead(au16data[0], 3) ;
  if (bitResult == 1) digitalWrite(DO3, LOW);
  else digitalWrite(DO3, HIGH);
}
//------- Digitale Eingänge Lesen und auf die DiscretInput-Register umschreiben ------
void read_digital_inputs()
{
  bool bitResult = digitalRead(DI0); // RX/TX
  bitWrite(au16data[1], 0, bitResult);
  bitResult = digitalRead(DI1);
  bitWrite(au16data[1], 1, bitResult);
  bitResult = digitalRead(DI2);
  bitWrite(au16data[1], 2, bitResult);
  bitResult = digitalRead(DI3);
  bitWrite(au16data[1], 3, bitResult);
}

//---------------- Analogwerte den Holding-Registern zuweisen -------------------------
//---------------- Analalogwerte umspeichern (floatReverse) -----------------------------
void value_mapping()
{
  values.floats[0] = power_fwd;   //umkopieren der Werte in den Transferbereich
  values.floats[1] = power_ref;
  values.floats[2] = power_drv;
  values.floats[3] = volt_50;
  values.floats[4] = amp_50;
  values.floats[5] = volt_15;
  values.floats[6] = SWR;
  values.floats[7] = temperature;
  values.floats[8] = 888.888;

  for (int i = 0; i <= 8; i++) {
    au16data[((i * 2)) + 2] = values.words[(i * 2) + 1]; //Daten aus den Modbusregistern in den Transferbereich umkopieren
    au16data[((i * 2) + 1) + 2] = values.words[(i * 2)]; //dabei erfolgt ein Bytetausch
  }
}
//-------------------Modbus-Slave nach einer Vorlage von [1] ---------------------------------
void modbus_slave() {
  state = slave.poll(au16data, 20);   // Status 0:
  // Status 1 bis 4: Kommunikationsfehler
  // Status größer 4: Daten korrekt
  if (state > 4) {
    u32wait = millis() + 50;
    digitalWrite(LED, HIGH );
  }
  if (millis() >  u32wait) digitalWrite(LED, LOW);
}
//-------------------------------------------------------------------------------------
